@@ -3,8 +3,6 @@
 # Licensed under the MIT License. See License.txt in the project root for license information.
 # --------------------------------------------------------------------------------------------
 
-import datetime
-import time
 import requests
 import adal
 
@@ -37,7 +35,7 @@ class AdalAuthentication(Authentication):  # pylint: disable=too-few-public-meth
         """
         external_tenant_tokens = None
         try:
-            scheme, token, full_token = self._token_retriever(sdk_resource)
+            scheme, token, token_entry = self._token_retriever(sdk_resource)
             if self._external_tenant_token_retriever:
                 external_tenant_tokens = self._external_tenant_token_retriever(sdk_resource)
         except CLIError as err:
@@ -45,48 +43,79 @@ class AdalAuthentication(Authentication):  # pylint: disable=too-few-public-meth
                 AdalAuthentication._log_hostname()
             raise err
         except adal.AdalError as err:
-            # pylint: disable=no-member
             if in_cloud_console():
                 AdalAuthentication._log_hostname()
-
-            err = (getattr(err, 'error_response', None) or {}).get('error_description') or str(err)
-            if 'AADSTS70008' in err:  # all errors starting with 70008 should be creds expiration related
-                raise CLIError("Credentials have expired due to inactivity. {}".format(
-                    "Please run 'az login'" if not in_cloud_console() else ''))
-            if 'AADSTS50079' in err:
-                raise CLIError("Configuration of your account was changed. {}".format(
-                    "Please run 'az login'" if not in_cloud_console() else ''))
-            if 'AADSTS50173' in err:
-                raise CLIError("The credential data used by CLI has been expired because you might have changed or "
-                               "reset the password. {}".format(
-                                   "Please clear browser's cookies and run 'az login'"
-                                   if not in_cloud_console() else ''))
-
-            raise CLIError(err)
+            adal_error_handler(err)
         except requests.exceptions.SSLError as err:
             from .util import SSLERROR_TEMPLATE
             raise CLIError(SSLERROR_TEMPLATE.format(str(err)))
         except requests.exceptions.ConnectionError as err:
             raise CLIError('Please ensure you have network connection. Error detail: ' + str(err))
 
-        return scheme, token, full_token, external_tenant_tokens
+        # scheme: str. The token scheme. Should always be 'Bearer'.
+        # token: str. The raw access token.
+        # token_entry: dict. The full token entry.
+        # external_tenant_tokens: [(scheme: str, token: str, token_entry: dict), ...]
+        return scheme, token, token_entry, external_tenant_tokens
+
+    def get_all_tokens(self, *scopes):
+        return self._get_token(_try_scopes_to_resource(scopes))
 
     # This method is exposed for Azure Core.
     def get_token(self, *scopes, **kwargs):  # pylint:disable=unused-argument
         logger.debug("AdalAuthentication.get_token invoked by Track 2 SDK with scopes=%s", scopes)
 
-        _, token, full_token, _ = self._get_token(_try_scopes_to_resource(scopes))
+        _, token, token_entry, _ = self._get_token(_try_scopes_to_resource(scopes))
 
-        try:
-            expires_on = full_token['expiresOn']
-            return AccessToken(token, int(datetime.datetime.strptime(expires_on, '%Y-%m-%d %H:%M:%S.%f').timestamp()))
-        except:  # pylint: disable=bare-except
-            pass  # To avoid crashes due to some unexpected token formats
+        # NEVER use expiresIn (expires_in) as the token is cached and expiresIn will be already out-of date
+        # when being retrieved.
 
-        try:
-            return AccessToken(token, int(full_token['expiresIn'] + time.time()))
-        except KeyError:  # needed to deal with differing unserialized MSI token payload
-            return AccessToken(token, int(full_token['expires_on']))
+        # User token entry sample:
+        # {
+        #     "tokenType": "Bearer",
+        #     "expiresOn": "2020-11-13 14:44:42.492318",
+        #     "resource": "https://management.core.windows.net/",
+        #     "userId": "test@azuresdkteam.onmicrosoft.com",
+        #     "accessToken": "eyJ0eXAiOiJKV...",
+        #     "refreshToken": "0.ATcAImuCVN...",
+        #     "_clientId": "04b07795-8ddb-461a-bbee-02f9e1bf7b46",
+        #     "_authority": "https://login.microsoftonline.com/54826b22-38d6-4fb2-bad9-b7b93a3e9c5a",
+        #     "isMRRT": True,
+        #     "expiresIn": 3599
+        # }
+
+        # Service Principal token entry sample:
+        # {
+        #     "tokenType": "Bearer",
+        #     "expiresIn": 3599,
+        #     "expiresOn": "2020-11-12 13:50:47.114324",
+        #     "resource": "https://management.core.windows.net/",
+        #     "accessToken": "eyJ0eXAiOiJKV...",
+        #     "isMRRT": True,
+        #     "_clientId": "22800c35-46c2-4210-b8a7-d8c3ec3b526f",
+        #     "_authority": "https://login.microsoftonline.com/54826b22-38d6-4fb2-bad9-b7b93a3e9c5a"
+        # }
+        if 'expiresOn' in token_entry:
+            import datetime
+            expires_on_timestamp = int(_timestamp(
+                datetime.datetime.strptime(token_entry['expiresOn'], '%Y-%m-%d %H:%M:%S.%f')))
+            return AccessToken(token, expires_on_timestamp)
+
+        # Cloud Shell (Managed Identity) token entry sample:
+        # {
+        #     "access_token": "eyJ0eXAiOiJKV...",
+        #     "refresh_token": "",
+        #     "expires_in": "2106",
+        #     "expires_on": "1605686811",
+        #     "not_before": "1605682911",
+        #     "resource": "https://management.core.windows.net/",
+        #     "token_type": "Bearer"
+        # }
+        if 'expires_on' in token_entry:
+            return AccessToken(token, int(token_entry['expires_on']))
+
+        from azure.cli.core.azclierror import CLIInternalError
+        raise CLIInternalError("No expiresOn or expires_on is available in the token entry.")
 
     # This method is exposed for msrest.
     def signed_session(self, session=None):  # pylint: disable=arguments-differ
@@ -118,6 +147,17 @@ class MSIAuthenticationWrapper(MSIAuthentication):
             # If available, use resource provided by SDK
             self.resource = resource
         self.set_token()
+        # Managed Identity token entry sample:
+        # {
+        #     "access_token": "eyJ0eXAiOiJKV...",
+        #     "client_id": "da95e381-d7ab-4fdc-8047-2457909c723b",
+        #     "expires_in": "86386",
+        #     "expires_on": "1605238724",
+        #     "ext_expires_in": "86399",
+        #     "not_before": "1605152024",
+        #     "resource": "https://management.azure.com/",
+        #     "token_type": "Bearer"
+        # }
         return AccessToken(self.token['access_token'], int(self.token['expires_on']))
 
     def set_token(self):
@@ -171,3 +211,52 @@ def _try_scopes_to_resource(scopes):
 
     # Exactly only one scope is provided
     return scopes_to_resource(scopes)
+
+
+class BasicTokenCredential:
+    # pylint:disable=too-few-public-methods
+    """A Track 2 implementation of msrest.authentication.BasicTokenAuthentication.
+    This credential shouldn't be used by any command module, expect azure-cli-core.
+    """
+    def __init__(self, access_token):
+        self.access_token = access_token
+
+    def get_token(self, *scopes, **kwargs):  # pylint:disable=unused-argument
+        # Because get_token can't refresh the access token, always mark the token as unexpired
+        import time
+        return AccessToken(self.access_token, int(time.time() + 3600))
+
+
+def _timestamp(dt):
+    # datetime.datetime can't be patched:
+    #   TypeError: can't set attributes of built-in/extension type 'datetime.datetime'
+    # So we wrap datetime.datetime.timestamp with this function.
+    # https://docs.python.org/3/library/unittest.mock-examples.html#partial-mocking
+    # https://williambert.online/2011/07/how-to-unit-testing-in-django-with-mocking-and-patching/
+    return dt.timestamp()
+
+
+def aad_error_handler(error: dict):
+    """ Handle the error from AAD server returned by ADAL or MSAL. """
+    login_message = ("To re-authenticate, please {}. If the problem persists, "
+                     "please contact your tenant administrator."
+                     .format("refresh Azure Portal" if in_cloud_console() else "run `az login`"))
+
+    # https://docs.microsoft.com/en-us/azure/active-directory/develop/reference-aadsts-error-codes
+    # Search for an error code at https://login.microsoftonline.com/error
+    msg = error.get('error_description')
+
+    from azure.cli.core.azclierror import AuthenticationError
+    raise AuthenticationError(msg, login_message)
+
+
+def adal_error_handler(err: adal.AdalError):
+    """ Handle AdalError. """
+    try:
+        aad_error_handler(err.error_response)
+    except AttributeError:
+        # In case of AdalError created as
+        #   AdalError('More than one token matches the criteria. The result is ambiguous.')
+        # https://github.com/Azure/azure-cli/issues/15320
+        from azure.cli.core.azclierror import UnknownError
+        raise UnknownError(str(err), recommendation="Please run `az account clear`, then `az login`.")
