@@ -127,8 +127,8 @@ def _prepare_client_kwargs_track2(cli_ctx):
     client_kwargs['logging_enable'] = True
 
     # Disable ARMHttpLoggingPolicy which logs only allowed headers
-    from azure.core.pipeline.policies import SansIOHTTPPolicy
-    client_kwargs['http_logging_policy'] = SansIOHTTPPolicy()
+    # from azure.core.pipeline.policies import SansIOHTTPPolicy
+    # client_kwargs['http_logging_policy'] = SansIOHTTPPolicy()
 
     # Prepare User-Agent header, used by UserAgentPolicy
     client_kwargs['user_agent'] = get_az_user_agent()
@@ -194,6 +194,16 @@ def _get_mgmt_service_client(cli_ctx,
         client_kwargs.update(_prepare_client_kwargs_track2(cli_ctx))
         client_kwargs['credential_scopes'] = resource_to_scopes(resource)
 
+        # Track 2 currently lacks the ability to take external credentials.
+        #   https://github.com/Azure/azure-sdk-for-python/issues/8313
+        # As a temporary workaround, manually add external tokens to 'x-ms-authorization-auxiliary' header.
+        #   https://docs.microsoft.com/en-us/azure/azure-resource-manager/management/authenticate-multi-tenant
+        if getattr(cred, "_external_tenant_token_retriever", None):
+            *_, external_tenant_tokens = cred.get_all_tokens(*resource_to_scopes(resource))
+            # Hard-code scheme to 'Bearer' as _BearerTokenCredentialPolicyBase._update_headers does.
+            client_kwargs['headers']['x-ms-authorization-auxiliary'] = \
+                ', '.join("Bearer {}".format(t[1]) for t in external_tenant_tokens)
+
     if subscription_bound:
         client = client_type(cred, subscription_id, **client_kwargs)
     else:
@@ -253,3 +263,6 @@ def _get_add_headers_callback(cli_ctx):
             pass
 
     return _add_headers
+
+
+prepare_client_kwargs_track2 = _prepare_client_kwargs_track2
